import os
import math
import shutil
import logging

from box import Box
import reusables
from kivy.uix.button import Button
from kivy.uix.image import Image
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.behaviors import ButtonBehavior
from kivy.uix.boxlayout import BoxLayout
from kivy.core.window import Window

from pyview.kivy_hover import MouseOver


log = reusables.get_logger('pyview', level=logging.DEBUG)


class PyViewError(Exception):
    """Something broke"""


class Selector(Button, MouseOver):
    """ Base class for Prev and Next Image Buttons"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.size_hint = (None, None)
        self.width = 100
        self.hpos = 99
        self.pos = (0, self.hpos)
        self.text = ">"
        self.opacity = 0

    def on_hover(self):
        self.opacity = .8

    def on_exit(self):
        self.opacity = 0


class NextButton(Selector):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.text = ">"

    def on_press(self):
        self.parent.image.next_image()


class PrevButton(Selector):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.text = "<"

    def on_press(self):
        self.parent.image.prev_image()


class PreviewBar(BoxLayout):
    """ Dynamic amount of preview images that can be selected"""

    def __init__(self, meta=None, **kwargs):
        self.meta = meta
        super().__init__(**kwargs)
        self.height = 100
        self.size_hint = (1, None)
        self.pos = (0, 0)
        self.images = 10
        self.update()

    def update(self):
        self.clear_widgets()
        for i in range(self.meta.pos - int(math.floor(self.images/2)),
                       self.meta.pos + int(math.ceil(self.images/2))):
            image_pos = i
            if image_pos < 0:
                image_pos = len(self.meta.list) + image_pos
            if image_pos >= len(self.meta.list):
                image_pos = image_pos - len(self.meta.list)
            img = MenuImage(source=os.path.join(self.meta.base,
                                                self.meta.list[image_pos].image),
                            image_pos=image_pos,
                            selected=True if image_pos == self.meta.pos else False)
            self.add_widget(img)

    def size_change(self, width=1920):
        self.images = int(width / 100)
        if self.images < 1:
            self.images = 1
        self.update()


class MenuImage(ButtonBehavior, Image, MouseOver):
    """ Preview image. The current one will always have full opacity, 
        otherwise only ones being hovered over will be full opacity. """

    def __init__(self, image_pos=0, selected=False, **kwargs):
        self.image_pos = image_pos
        self.selected = selected
        super().__init__(**kwargs)
        self.width = 100
        if not self.selected:
            self.opacity = 0.6
        self.size_hint = None, None

    def on_press(self):
        if not self.selected:
            self.parent.parent.change_to_image(self.image_pos)

    def on_hover(self):
        if not self.selected:
            self.opacity = 1.0

    def on_exit(self):
        if not self.selected:
            self.opacity = 0.6


class MainImage(Image):

    def __init__(self, meta=None, preview_bar=None, **kwargs):
        self.meta = meta
        if not preview_bar:
            raise PyViewError("Preview_bar must be provided")
        self.preview_bar = preview_bar
        super().__init__(source=self.gen_image(), **kwargs)
        self.pos = (0, 100)
        self.size_hint_x = 1
        self.size_hint_y = None

    def gen_image(self):
        return os.path.join(self.meta.base,
                            self.meta.list[self.meta.pos].image)

    def next_image(self):
        self.meta.pos += 1
        if self.meta.pos >= len(self.meta.list):
            self.meta.pos = 0
        self.source = self.gen_image()
        self.reload()
        self.preview_bar.update()

    def prev_image(self):
        self.meta.pos -= 1
        if self.meta.pos < 0:
            self.meta.pos = len(self.meta.list) - 1
        self.source = self.gen_image()
        self.reload()
        self.preview_bar.update()


class ImageViewer(FloatLayout):

    def __init__(self, base_dir=r"/home/james/Pictures",
                 delete_dir=r"/home/james/delete", **kwargs):
        super().__init__(**kwargs)
        # Capture keyboard input
        self._keyboard = Window.request_keyboard(self._keyboard_closed, self)
        self._keyboard.bind(on_key_down=self._on_keyboard_down)
        # Make ImageViewer object take up entire parent space
        self.pos = (0, 0)
        self.size_hint = (1, 1)
        # Meta data about the image set shared between main image and previews
        self.image_data = Box()
        self.image_data.del_dir = delete_dir
        os.makedirs(self.image_data.del_dir, exist_ok=True)
        self.image_data.base = base_dir
        self.image_data.pos = 0
        self.image_data.list = []
        self._get_images()
        self.image_data.list.sort(key=lambda x: x['created'], reverse=True)
        # Define widgets used so we can reference them elsewhere
        self.preview = PreviewBar(meta=self.image_data)
        self.image = MainImage(meta=self.image_data, preview_bar=self.preview)
        self.next_button = NextButton()
        self.prev_button = PrevButton()

        self.add_widget(self.image, index=2)
        self.add_widget(self.preview, index=0)
        self.add_widget(self.next_button, index=1)
        self.add_widget(self.prev_button, index=1)

    def _get_images(self):
        for x in os.listdir(self.image_data.base):
            if x.lower().endswith(reusables.exts.pictures):
                try:
                    created_time = os.path.getctime(os.path.join(
                        self.image_data.base, x))
                except OSError:
                    created_time = 0
                data = {'image': x, 'created': created_time}
                self.image_data.list.append(data)

    def on_size(self, obj, size):
        """Make sure all children sizes adjust properly"""
        log.debug(f"Resizing image window to {size[0]}x{size[1]}")
        self.next_button.pos = (self.width - 99, self.next_button.hpos)
        self.preview.size_change(width=size[0])
        self.image.height = self.height - 100
        self.next_button.height = self.height - 100
        self.prev_button.height = self.height - 100

    def delete_image(self):
        img = self.image_data.list[self.image_data.pos]
        self.image_data.list.remove(img)
        del_name = (f"{reusables.now().isoformat().replace(':', '.')}"
                    f"-_-{img.image}")
        shutil.move(os.path.join(self.image_data.base, img.image),
                    os.path.join(self.image_data.del_dir, del_name))
        self.change_to_image(self.image_data.pos)

    def update_preview(self):
        self.preview.update()

    def change_to_image(self, image_pos):
        self.image_data.pos = image_pos
        self.image.source = self.image.gen_image()
        self.image.reload()
        self.preview.update()

    def _keyboard_closed(self):
        self._keyboard.unbind(on_key_down=self._on_keyboard_down)
        self._keyboard = None

    def _on_keyboard_down(self, keyboard, keycode, text, modifiers):
        if keycode[1] == 'right':
            self.image.next_image()
        elif keycode[1] == 'left':
            self.image.prev_image()
        elif keycode[1] == 'delete':
            self.delete_image()
        return True
